

/* field.c */


/*------------------------------------------------------------------------
 * Module for field arithmetics
 *------------------------------------------------------------------------
 * This version of field is part of the Solitaire Brute Force tool.
 * Copyright (C) 2001 Holger Hasselbach <H.Hasselbach@galad.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *------------------------------------------------------------------------
 */


/*----------------------------------------------------------------------*/

#include "field.h"
#include "mymemory.h"
#include "mystring.h"

#include <assert.h>

/*----------------------------------------------------------------------*/


/*------------------------------------------------------------------------
 * FIELDVAL
 *------------------------------------------------------------------------
 */


/*
 * To allow the calculation of none-symmetric start fields, the normalization
 * is disabled with these fields. To reduce memory usage and calculation time
 * in this case, the changes were such distributed over the field module that
 * it was much more easy to introduce a file scope global variable. It is set
 * in FieldValInitStart only, and it controls (when it is FALSE):
 * - FieldHsh uses only one counter for the ways in the full tree.
 * - Both All and Nrm statistics give the same results (the Stat module was
 *   not modified).
 * - The FieldSttCalculate function sets the SymType for any field to s1, the
 *   fields are not normalized, and any following field is reached by only
 *   1 move from the actual field.
 */

static tBOOL b_FieldVal_StartIsS1 = FALSE;

/*---------------------------------------------*/

#define cn_FieldVal_Counter 2
#define cn_FieldVal_CntAll  0
#define cn_FieldVal_CntNrm  1


struct s_FIELDVAL
{
    tULONG     ulField;
    tULONG     ulBit32;

    tpsWORDCNT psCounter;
    tCOUNT     cnMoves;
};

typedef struct s_FIELDVAL tsFIELDVAL;

/*---------------------------------------------*/

static tULONG ul_FieldVal_SymNormal      (tULONG ulField);
static tCOUNT cn_FieldVal_SymType        (tULONG ulField);
static tCOUNT cn_FieldVal_SymTypeToIndex (tCOUNT cnSymType);

static tVOID  v_FieldVal_HashRestore     (tpsFIELDVAL psVal, tCOUNT cnHashKey);

static tCOUNT cn_FieldVal_Stones         (tpsFIELDVAL psVal);
static tVOID  v_FieldVal_StrToPattern    (tpsFIELDVAL psVal, tPSZc pzStr);

/*---------------------------------------------*/

static tVOID v_FieldVal_MoveLoopInit (tpCOUNT pcnIndex);
static tBOOL b_FieldVal_MoveLoopNext (tpCOUNT pcnIndex, tULONG ulField, tULONG ulBit32, tpULONG pulNewField, tpULONG pulNewBit32);



/*------------------------------------------------------------------------
 * FIELDHSH
 *------------------------------------------------------------------------
 */


struct s_FIELDHSH
{
    tCOUNT    cnHashKey;

    tULONG    aulFields[cnFieldHashEntries];
    tsWORDCNT sCountAll;
    tsWORDCNT sCountNrm;
};

typedef struct s_FIELDHSH tsFIELDHSH;



/*------------------------------------------------------------------------
 * FIELDSTT
 *------------------------------------------------------------------------
 */


struct s_FIELDSTT
{
    tULONG     ulField;
    tULONG     ulBit32;
    tpsWORDCNT psCounter;

    tCOUNT     cnStones;
    tCOUNT     cnSymType;
    tCOUNT     cnSymIndex;
    tBOOL      bIsWinning;

    tCOUNT     cnNextNum;                        /* Number of following normalized fields   */
    tULONG     aulNextField[cnFieldNumberMoves]; /* Following field values                  */
    tULONG     aulNextBit32[cnFieldNumberMoves];
    tCOUNT     acnNextMoves[cnFieldNumberMoves]; /* Number of moves reaching the same field */
    tCOUNT     cnNextLoop;                       /* Loop counter for evaluating next fields */

    tCOUNT     cnMovesAll;                       /* cnMovesNrm == cnNextNum */
};

typedef struct s_FIELDSTT tsFIELDSTT;


/*----------------------------------------------------------------------*/



/*------------------------------------------------------------------------
 * FIELDVAL
 *------------------------------------------------------------------------
 */


/*-----------------------------------------------
 * Create a new FIELDVAL
 *-----------------------------------------------
 */

tpsFIELDVAL psFieldValCreate (tVOID)
{
    tpsFIELDVAL psVal;

    psVal = pvMemoryAllocObject(sFIELDVAL);

    psVal->psCounter = psWordCntCreate(cn_FieldVal_Counter);

    return (psVal);
}


/*-----------------------------------------------
 * Destroy a FIELDVAL
 *-----------------------------------------------
 */

tpsFIELDVAL psFieldValDestroy (tpsFIELDVAL psVal)
{
    if (psVal)
      {
        psWordCntDestroy(psVal->psCounter);

        pvMemoryFree(psVal);
      }

    return (NULL);
}


/*-----------------------------------------------
 * Memory statistic
 *-----------------------------------------------
 */

tpsFIELDVAL psFieldValMemUsage (tpsFIELDVAL psVal, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    if (psVal)
      {
        psWordCntMemUsage(psVal->psCounter,pcnSizeof,pcnBlocks);

        *pcnSizeof += cnMemorySizeofObject(sFIELDVAL);
        *pcnBlocks += 1;
      }

    return (psVal);
}


/*---------------------------------------------------------
 * Symmetry operations
 *---------------------------------------------------------
 *       _0 _1 _2
 *       12 13 14
 * 11 23 24 25 26 15 _3
 * 10 22 31 32 27 16 _4
 * _9 21 30 29 28 17 _5
 *       20 19 18
 *       _8 _7 _6
 *
 * All operations are independent of bit 32.
 *---------------------------------------------------------
 */


/*-----------------------------------------------
 * Macro: Turn field 90 degrees right
 *-----------------------------------------------
 */


/* static tULONG ul_FieldVal_SymTurn (tULONG ulField)
{
    return ( ((ulField & 0x001ff1ff) << 3) | ((ulField & 0x00e00e00) >> 9) |
             ((ulField & 0x3f000000) << 2) | ((ulField & 0xc0000000) >> 6)
           );
} */


#define ul_FieldVal_SymTurn(ulField)                                          \
            ( ((ulField & 0x001ff1ff) << 3) | ((ulField & 0x00e00e00) >> 9) | \
              ((ulField & 0x3f000000) << 2) | ((ulField & 0xc0000000) >> 6)   \
            )


/*-----------------------------------------------
 * Macro: Mirror field on vertical axis
 *-----------------------------------------------
 */

/* static tULONG ul_FieldVal_SymMirror (tULONG ulField)
{
    return ( ((ulField & 0x11041041) << 2) | ((ulField & 0x44104104) >> 2) |
             ((ulField & 0x00008008) << 8) | ((ulField & 0x00800800) >> 8) |
             ((ulField & 0x00010010) << 6) | ((ulField & 0x00400400) >> 6) |
             ((ulField & 0x08020020) << 4) | ((ulField & 0x80200200) >> 4) |
              (ulField & 0x22082082)
           );
} */


#define ul_FieldVal_SymMirror(ulField)                                        \
            ( ((ulField & 0x11041041) << 2) | ((ulField & 0x44104104) >> 2) | \
              ((ulField & 0x00008008) << 8) | ((ulField & 0x00800800) >> 8) | \
              ((ulField & 0x00010010) << 6) | ((ulField & 0x00400400) >> 6) | \
              ((ulField & 0x08020020) << 4) | ((ulField & 0x80200200) >> 4) | \
               (ulField & 0x22082082)                                         \
            )


/*-----------------------------------------------
 * Local: Calculate symmetric integer normal (== max value)
 *-----------------------------------------------
 */

static tULONG ul_FieldVal_SymNormal (tULONG ulField)
{
    tULONG ulMax = ulField;

    if ( (ulField = ul_FieldVal_SymTurn  (ulField)) > ulMax ) ulMax = ulField;
    if ( (ulField = ul_FieldVal_SymTurn  (ulField)) > ulMax ) ulMax = ulField;
    if ( (ulField = ul_FieldVal_SymTurn  (ulField)) > ulMax ) ulMax = ulField;
    if ( (ulField = ul_FieldVal_SymMirror(ulField)) > ulMax ) ulMax = ulField;
    if ( (ulField = ul_FieldVal_SymTurn  (ulField)) > ulMax ) ulMax = ulField;
    if ( (ulField = ul_FieldVal_SymTurn  (ulField)) > ulMax ) ulMax = ulField;
    if ( (ulField = ul_FieldVal_SymTurn  (ulField)) > ulMax ) ulMax = ulField;

    return (ulMax);
}


/*-----------------------------------------------
 * Local: Get the symmetry type
 *-----------------------------------------------
 * None-symmetric: cnC = 1 ==> 8 different fields
 * 4th -symmetric: cnC = 2 ==> 4 different fields
 * Half-symmetric: cnC = 4 ==> 2 different fields
 * Full-symmetric: cnC = 8 ==> 1 field
 *-----------------------------------------------
 */

static tCOUNT cn_FieldVal_SymType (tULONG ulField)
{
    tULONG ulF = ulField;
    tCOUNT cnC = 1;

    if ( (ulField = ul_FieldVal_SymTurn  (ulField)) == ulF ) cnC++;
    if ( (ulField = ul_FieldVal_SymTurn  (ulField)) == ulF ) cnC++;
    if ( (ulField = ul_FieldVal_SymTurn  (ulField)) == ulF ) cnC++;
    if ( (ulField = ul_FieldVal_SymMirror(ulField)) == ulF ) cnC++;
    if ( (ulField = ul_FieldVal_SymTurn  (ulField)) == ulF ) cnC++;
    if ( (ulField = ul_FieldVal_SymTurn  (ulField)) == ulF ) cnC++;
    if ( (ulField = ul_FieldVal_SymTurn  (ulField)) == ulF ) cnC++;

    return ( 8 / cnC );
}


/*-----------------------------------------------
 * Local: Calculate SymIndex from SymType
 *-----------------------------------------------
 * SymType :  1  2  4  8
 * SymIndex:  0  1  2  3
 *-----------------------------------------------
 */

                                           /* 0,1,2,3,4,5,6,7,8 */
static tCOUNT acn_FieldVal_TypeToIndex[9] = { 0,0,1,0,2,0,0,0,3 };

static tCOUNT cn_FieldVal_SymTypeToIndex (tCOUNT cnSymType)
{
    assert(cnSymType < 9);

    return ( acn_FieldVal_TypeToIndex[cnSymType] );
}


/*-----------------------------------------------
 * Calculate SymType from SymIndex
 *-----------------------------------------------
 */

tCOUNT cnFieldValSymIndexToType (tCOUNT cnSymIndex)
{
    assert(cnSymIndex < cnFieldSymIndexRange);

    return (1 << cnSymIndex);
}


/*---------------------------------------------------------
 * HashKey operations
 *---------------------------------------------------------
 * The central 33rd bit is stored in the hash key itself.
 * For other Solitaire patterns, more bits may be stored
 * there to avoid changes in FIELDHSH.
 *
 * The hash calculation makes no difference on the field
 * pattern, all 2^33 field values are valid input parameters.
 * The function is in no way optimized for equal distribution
 * of keys, but some samples tested by eyesight were quite
 * good (at least enough).
 *
 * When the range is changed, a power of 2 is recommended for
 * symmetry of value usage when bits are stored in the hash
 * key. Because of the different numbers necessary for the
 * calculation, it is IMO of no value to use constants other
 * than to obfuscate the code. All numbers are hard coded.
 *---------------------------------------------------------
 */


/*-----------------------------------------------
 * Calculate the hash key for a field
 *-----------------------------------------------
 */

tCOUNT cnFieldValHashKey (tpsFIELDVAL psVal)
{
    tCOUNT cnHash;

    cnHash = ((psVal->ulField      ) & 4095) +
             ((psVal->ulField >> 12) & 4095) +
             ((psVal->ulField >> 24) & 4095);

    while (cnHash > 2047)
        cnHash = (cnHash & 2047) + (cnHash >> 11);

    return ( cnHash | ((psVal->ulBit32 & 1) << 11) );
}


/*-----------------------------------------------
 * Local: Restore the hash key coded bits
 *-----------------------------------------------
 */

static tVOID v_FieldVal_HashRestore (tpsFIELDVAL psVal, tCOUNT cnHashKey)
{
    psVal->ulBit32 = (cnHashKey >> 11) & 1;
}



/*---------------------------------------------------------
 * Other pattern related functions
 *---------------------------------------------------------
 */


/*-----------------------------------------------
 * Initialize start field
 *-----------------------------------------------
 * pzPattern must contain at least StonesRange-1
 * chars of '0' or '1', else it is ignored. When
 * it is valid, it is converted from top to bottom
 * and from left to right.
 *-----------------------------------------------
 */

tVOID vFieldValInitStart (tpsFIELDVAL psVal, tPSZc pzPattern)
{
    psVal->ulField = 0xffffffff;
    psVal->ulBit32 = 0;

    /*
    psVal->ulField = 0x00492492;
    psVal->ulBit32 = 0;
    */

    v_FieldVal_StrToPattern(psVal,pzPattern);

    vWordCntInit1(psVal->psCounter,cn_FieldVal_CntAll);
    vWordCntInit1(psVal->psCounter,cn_FieldVal_CntNrm);

    psVal->cnMoves = 1;

    b_FieldVal_StartIsS1 = (cn_FieldVal_SymType(psVal->ulField) == 1);
}


/*-----------------------------------------------
 * Get StartIsS1-flag
 *-----------------------------------------------
 */

tBOOL bFieldValStartIsS1 (tVOID)
{
    return (b_FieldVal_StartIsS1);
}


/*-----------------------------------------------
 * Test field for winning end pattern
 *-----------------------------------------------
 */

tBOOL bFieldValIsWinning (tpsFIELDVAL psVal)
{
    return ( (psVal->ulField == 0) && (psVal->ulBit32 == 1) );
}


/*-----------------------------------------------
 * Local: Get the number of stones on the field
 *-----------------------------------------------
 */

static tUBYTE aub_FieldVal_NibBits[16] =
  {
    0, /* 0000 */
    1, /* 0001 */
    1, /* 0010 */
    2, /* 0011 */
    1, /* 0100 */
    2, /* 0101 */
    2, /* 0110 */
    3, /* 0111 */
    1, /* 1000 */
    2, /* 1001 */
    2, /* 1010 */
    3, /* 1011 */
    2, /* 1100 */
    3, /* 1101 */
    3, /* 1110 */
    4  /* 1111 */
  };

static tCOUNT cn_FieldVal_Stones (tpsFIELDVAL psVal)
{
    return ( aub_FieldVal_NibBits[(psVal->ulField      ) & 15] +
             aub_FieldVal_NibBits[(psVal->ulField >>  4) & 15] +
             aub_FieldVal_NibBits[(psVal->ulField >>  8) & 15] +
             aub_FieldVal_NibBits[(psVal->ulField >> 12) & 15] +
             aub_FieldVal_NibBits[(psVal->ulField >> 16) & 15] +
             aub_FieldVal_NibBits[(psVal->ulField >> 20) & 15] +
             aub_FieldVal_NibBits[(psVal->ulField >> 24) & 15] +
             aub_FieldVal_NibBits[(psVal->ulField >> 28) & 15] +
             (psVal->ulBit32 & 1)
           );
}


/*-----------------------------------------------
 * Build or analyse field pattern string
 *-----------------------------------------------
 * 4 chars are used:
 * '0'  for a 0 position (empty)
 * '1'  for a 1 position (stone)
 * ' '  for an empty corner
 * '\n' for a line feed
 *-----------------------------------------------
 */

#define cn_FieldVal_PatStrLen 57

static tUBYTE aub_FieldVal_PatStrBit[cn_FieldVal_PatStrLen] =
  {
     64,64, 0, 1, 2,64,64,65,
     64,64,12,13,14,64,64,65,
     11,23,24,25,26,15, 3,65,
     10,22,31,32,27,16, 4,65,
      9,21,30,29,28,17, 5,65,
     64,64,20,19,18,64,64,65,
     64,64, 8, 7, 6,64,64,65,
    128
  };

/*---------------------------------------------*/

tpPSZ ppzFieldValPatternToStr (tpPSZ ppzStr, tpsFIELDVAL psVal)
{
    tCOUNT cnI;
    tUBYTE ubBit;
    tCHAR  cChr;

    ppzStringReserve(ppzStr,cn_FieldVal_PatStrLen);

    for (cnI = 0; cnI < cn_FieldVal_PatStrLen; cnI++)
      {
        ubBit = aub_FieldVal_PatStrBit[cnI];
        switch (ubBit)
          {
          case  64 : cChr = ' '; break;
          case  65 :
          case 128 : cChr = '\n'; break;
          case  32 : cChr = ( psVal->ulBit32 & 1 ? '1' : '0' ); break;
          default  : cChr = ( psVal->ulField & ((tULONG)1 << ubBit) ? '1' : '0' ); break;
          }
        (*ppzStr)[cnI] = cChr;
      }

    return (ppzStr);
}

/*---------------------------------------------*/

static tVOID v_FieldVal_StrToPattern (tpsFIELDVAL psVal, tPSZc pzStr)
{
    tpUBYTE pubPat;
    tCOUNT  cnIBit;
    tULONG  ulField,ulBit32;

    if (pzStr)
      {
        ulField =
        ulBit32 = 0;
        pubPat  = aub_FieldVal_PatStrBit;
        cnIBit  = 0;
        while (cnIBit < (cnFieldStonesRange - 1))
          {
            while ( *pzStr && (*pzStr != '0') && (*pzStr != '1') )
                pzStr++;
            while ( (*pubPat != 128) && (*pubPat >= 64) )
                pubPat++;

            if ( *pzStr && (*pubPat != 128) )
              {
                if (*pubPat < 32)
                    ulField |= (tULONG)(*pzStr - '0') << *pubPat;
                else
                    ulBit32 |= (tULONG)(*pzStr - '0') << (*pubPat - 32);

                pzStr++;
                pubPat++;
                cnIBit++;
              }
            else
                break;
          }

        if (cnIBit == (cnFieldStonesRange - 1))
          {
            psVal->ulField = ulField;
            psVal->ulBit32 = ulBit32;
          }
      }
}



/*---------------------------------------------------------
 * Moves
 *---------------------------------------------------------
 */


/*-----------------------------------------------
 * Move possibilities
 *-----------------------------------------------
 *       _0 _1 _2
 *       12 13 14
 * 11 23 24 25 26 15 _3
 * 10 22 31 32 27 16 _4
 * _9 21 30 29 28 17 _5
 *       20 19 18
 *       _8 _7 _6
 *-----------------------------------------------
 */

struct s_FIELDVALMOVE
{
    tUBYTE ubJump;
    tUBYTE ubOver;
    tUBYTE ubLand;
};

typedef struct s_FIELDVALMOVE ts_FIELDVALMOVE;
typedef ts_FIELDVALMOVE       *tps_FIELDVALMOVE;

static ts_FIELDVALMOVE as_FieldVal_MoveTab[cnFieldNumberMoves] =
  {
    {24,12, 0} , {25,13, 1} , {26,14, 2} , /* Up */
    {31,24,12} , {32,25,13} , {27,26,14} ,
    {30,31,24} , {29,32,25} , {28,27,26} ,
    {20,30,31} , {19,29,32} , {18,28,27} ,
    { 8,20,30} , { 7,19,29} , { 6,18,28} ,
    { 9,10,11} , {21,22,23} , {17,16,15} , { 5, 4, 3} ,

    {30,20, 8} , {29,19, 7} , {28,18, 6} , /* Down */
    {31,30,20} , {32,29,19} , {27,28,18} ,
    {24,31,30} , {25,32,29} , {26,27,28} ,
    {12,24,31} , {13,25,32} , {14,26,27} ,
    { 0,12,24} , { 1,13,25} , { 2,14,26} ,
    {11,10, 9} , {23,22,21} , {15,16,17} , { 3, 4, 5} ,

    {24,23,11} , {31,22,10} , {30,21, 9} , /* Left */
    {25,24,23} , {32,31,22} , {29,30,21} ,
    {26,25,24} , {27,32,31} , {28,29,30} ,
    {15,26,25} , {16,27,32} , {17,28,29} ,
    { 3,15,26} , { 4,16,27} , { 5,17,28} ,
    { 2, 1, 0} , {14,13,12} , {18,19,20} , { 6, 7, 8} ,

    {26,15, 3} , {27,16, 4} , {28,17, 5} , /* Right */
    {25,26,15} , {32,27,16} , {29,28,17} ,
    {24,25,26} , {31,32,27} , {30,29,28} ,
    {23,24,25} , {22,31,32} , {21,30,29} ,
    {11,23,24} , {10,22,31} , { 9,21,30} ,
    { 0, 1, 2} , {12,13,14} , {20,19,18} , { 8, 7, 6}
  };


/*-----------------------------------------------
 * Local: Initialize loop for new field search
 *-----------------------------------------------
 */

static tVOID v_FieldVal_MoveLoopInit (tpCOUNT pcnIndex)
{
    *pcnIndex = 0;
}


/*-----------------------------------------------
 * Local: Search the next possible move
 *-----------------------------------------------
 * Returns FALSE when no more moves possible.
 *-----------------------------------------------
 */

#define b_FieldVal_BitTest(ubBit) ( ( (ubBit) > 31 ? ulBit32 : ulField >> (ubBit) ) & 1 )

static tBOOL b_FieldVal_MoveLoopNext (tpCOUNT pcnIndex, tULONG ulField, tULONG ulBit32,
                                                        tpULONG pulNewField, tpULONG pulNewBit32)
{
    tBOOL            bFound = FALSE;
    tps_FIELDVALMOVE psMove;

    while ( !bFound && (*pcnIndex < cnFieldNumberMoves) )
      {
        psMove = &(as_FieldVal_MoveTab[*pcnIndex]);

        if (  b_FieldVal_BitTest(psMove->ubJump) &&
              b_FieldVal_BitTest(psMove->ubOver) &&
             !b_FieldVal_BitTest(psMove->ubLand) )
          {
            *pulNewField = ulField;
            *pulNewBit32 = ulBit32;

            if (psMove->ubJump > 31)
                *pulNewBit32 = 0;
            else
                *pulNewField &= ~((tULONG)1 << psMove->ubJump);

            if (psMove->ubOver > 31)
                *pulNewBit32 = 0;
            else
                *pulNewField &= ~((tULONG)1 << psMove->ubOver);

            if (psMove->ubLand > 31)
                *pulNewBit32 = 1;
            else
                *pulNewField |=  ((tULONG)1 << psMove->ubLand);

            bFound = TRUE;
          }

        (*pcnIndex)++;
      }

    return (bFound);
}



/*------------------------------------------------------------------------
 * FIELDHSH
 *------------------------------------------------------------------------
 */


/*-----------------------------------------------
 * Create a new FIELDHSH
 *-----------------------------------------------
 */

tpsFIELDHSH psFieldHshCreate (tVOID)
{
    tpsFIELDHSH psHsh;

    psHsh = pvMemoryAllocObject(sFIELDHSH);

    psWordCntInit(&(psHsh->sCountAll),( b_FieldVal_StartIsS1 ? cnFieldHashEntries : 0 ));
    psWordCntInit(&(psHsh->sCountNrm),cnFieldHashEntries);

    return (psHsh);
}


/*-----------------------------------------------
 * Destroy a FIELDHSH
 *-----------------------------------------------
 */

tpsFIELDHSH psFieldHshDestroy (tpsFIELDHSH psHsh)
{
    if (psHsh)
      {
        psWordCntExit(&(psHsh->sCountAll));
        psWordCntExit(&(psHsh->sCountNrm));

        pvMemoryFree(psHsh);
      }

    return (NULL);
}


/*-----------------------------------------------
 * Memory statistics
 *-----------------------------------------------
 */

tpsFIELDHSH psFieldHshMemUsage (tpsFIELDHSH psHsh, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    if (psHsh)
      {
        psWordCntMemInner(&(psHsh->sCountAll),pcnSizeof,pcnBlocks);
        psWordCntMemInner(&(psHsh->sCountNrm),pcnSizeof,pcnBlocks);

        *pcnSizeof += cnMemorySizeofObject(sFIELDHSH);
        *pcnBlocks += 1;
      }

    return (psHsh);
}


/*-----------------------------------------------
 * Search field in FIELDHSH
 *-----------------------------------------------
 * Returns TRUE with cnIndex if found
 * else FALSE with cnIndex unchanged.
 *-----------------------------------------------
 */

tBOOL bFieldHshSearch (tpsFIELDHSH psHsh, tCOUNT cnEntries, tpsFIELDVAL psVal, tpCOUNT pcnIndex)
{
    tCOUNT  cnI;
    tpULONG pulF;
    tBOOL   bFound = FALSE;

    assert(cnEntries <= cnFieldHashEntries);

    pulF = psHsh->aulFields;
    for (cnI = 0; cnI < cnEntries; cnI++, pulF++)
        if (psVal->ulField == *pulF)
          {
            *pcnIndex = cnI;
            bFound    = TRUE;
            break;
          }

    return (bFound);
}


/*-----------------------------------------------
 * Set a field in the FIELDHSH
 *-----------------------------------------------
 */

tVOID vFieldHshSetField (tpsFIELDHSH psHsh, tCOUNT cnIndex, tpsFIELDVAL psVal)
{
    assert(cnIndex < cnFieldHashEntries);

    psHsh->aulFields[cnIndex] = psVal->ulField;

    if (b_FieldVal_StartIsS1)
      {
        vWordCntInit0(&(psHsh->sCountAll),cnIndex);
        vWordCntAddMulti(&(psHsh->sCountAll),cnIndex,psVal->psCounter,cn_FieldVal_CntAll,psVal->cnMoves);
      }

    vWordCntCopy(&(psHsh->sCountNrm),cnIndex,psVal->psCounter,cn_FieldVal_CntNrm);
}


/*-----------------------------------------------
 * Get a field from the FIELDHSH
 *-----------------------------------------------
 */

tVOID vFieldHshGetField (tpsFIELDHSH psHsh, tCOUNT cnIndex, tpsFIELDVAL psVal)
{
    assert(cnIndex < cnFieldHashEntries);

    v_FieldVal_HashRestore(psVal,psHsh->cnHashKey);
    psVal->ulField = psHsh->aulFields[cnIndex];

    vWordCntCopy(psVal->psCounter,cn_FieldVal_CntAll,
                 ( b_FieldVal_StartIsS1 ? &(psHsh->sCountAll) : &(psHsh->sCountNrm) ),cnIndex);
    vWordCntCopy(psVal->psCounter,cn_FieldVal_CntNrm,&(psHsh->sCountNrm),cnIndex);
    psVal->cnMoves = 1;
}


/*-----------------------------------------------
 * Add a field to an existing field in the FIELDHSH
 *-----------------------------------------------
 */

tVOID vFieldHshAddField (tpsFIELDHSH psHsh, tCOUNT cnIndex, tpsFIELDVAL psVal)
{
    assert(cnIndex < cnFieldHashEntries);

    if (b_FieldVal_StartIsS1)
        vWordCntAddMulti(&(psHsh->sCountAll),cnIndex,psVal->psCounter,cn_FieldVal_CntAll,psVal->cnMoves);
    vWordCntAdd(&(psHsh->sCountNrm),cnIndex,psVal->psCounter,cn_FieldVal_CntNrm);
}


/*-----------------------------------------------
 * Set the hash key
 *-----------------------------------------------
 */

tVOID vFieldHshSetKey (tpsFIELDHSH psHsh, tCOUNT cnHashKey)
{
    psHsh->cnHashKey = cnHashKey;
}


/*-----------------------------------------------
 * Get the hash key
 *-----------------------------------------------
 */

tCOUNT cnFieldHshGetKey (tpsFIELDHSH psHsh)
{
    return (psHsh->cnHashKey);
}



/*------------------------------------------------------------------------
 * FIELDSTT
 *------------------------------------------------------------------------
 */


/*-----------------------------------------------
 * Create a new FIELDSTT
 *-----------------------------------------------
 */

tpsFIELDSTT psFieldSttCreate (tVOID)
{
    tpsFIELDSTT psStt;

    psStt = pvMemoryAllocObject(sFIELDSTT);

    psStt->psCounter = psWordCntCreate(cn_FieldVal_Counter);

    return (psStt);
}


/*-----------------------------------------------
 * Destroy a FIELDSTT
 *-----------------------------------------------
 */

tpsFIELDSTT psFieldSttDestroy (tpsFIELDSTT psStt)
{
    if (psStt)
      {
        psWordCntDestroy(psStt->psCounter);
        pvMemoryFree(psStt);
      }

    return (NULL);
}


/*-----------------------------------------------
 * Memory statistics
 *-----------------------------------------------
 */

tpsFIELDSTT psFieldSttMemUsage (tpsFIELDSTT psStt, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    if (psStt)
      {
        psWordCntMemUsage(psStt->psCounter,pcnSizeof,pcnBlocks);

        *pcnSizeof += cnMemorySizeofObject(sFIELDSTT);
        *pcnBlocks += 1;
      }

    return (psStt);
}


/*-----------------------------------------------
 * Calculate the field statistic
 *-----------------------------------------------
 */

tVOID vFieldSttCalculate (tpsFIELDSTT psStt, tpsFIELDVAL psVal)
{
    tCOUNT  cnI,cnIndex;
    tULONG  ulNewField,ulNewBit32;
    tpULONG pulNextF,pulNextB;

    psStt->ulField = psVal->ulField;
    psStt->ulBit32 = psVal->ulBit32;
    vWordCntCopy(psStt->psCounter,cn_FieldVal_CntAll,psVal->psCounter,cn_FieldVal_CntAll);
    vWordCntCopy(psStt->psCounter,cn_FieldVal_CntNrm,psVal->psCounter,cn_FieldVal_CntNrm);

    psStt->cnStones   = cn_FieldVal_Stones(psVal);
    psStt->cnSymType  = ( b_FieldVal_StartIsS1 ? cn_FieldVal_SymType(psVal->ulField) : 1 );
    psStt->cnSymIndex = cn_FieldVal_SymTypeToIndex(psStt->cnSymType);
    psStt->bIsWinning = bFieldValIsWinning(psVal);

    psStt->cnNextNum  =
    psStt->cnMovesAll = 0;
    v_FieldVal_MoveLoopInit(&cnIndex);
    while ( b_FieldVal_MoveLoopNext(&cnIndex,psVal->ulField,psVal->ulBit32,&ulNewField,&ulNewBit32) )
      {
        if (b_FieldVal_StartIsS1)
          {
            ulNewField = ul_FieldVal_SymNormal(ulNewField);

            pulNextF = psStt->aulNextField;
            pulNextB = psStt->aulNextBit32;
            for (cnI = 0; cnI < psStt->cnNextNum; cnI++, pulNextF++, pulNextB++)
                if ( (ulNewField == *pulNextF) && (ulNewBit32 == *pulNextB) )
                  {
                    (psStt->acnNextMoves[cnI])++;
                    break;
                  }
          }
        else
          {
            cnI      = psStt->cnNextNum;
            pulNextF = &(psStt->aulNextField[cnI]);
            pulNextB = &(psStt->aulNextBit32[cnI]);
          }

        if (cnI == psStt->cnNextNum)
          {
            *pulNextF = ulNewField;
            *pulNextB = ulNewBit32;
            psStt->acnNextMoves[cnI] = 1;
            psStt->cnNextNum++;
          }

        psStt->cnMovesAll++;
      }
}


/*-----------------------------------------------
 * Initialize move loop
 *-----------------------------------------------
 */

tVOID vFieldSttLoopInit (tpsFIELDSTT psStt)
{
    psStt->cnNextLoop = 0;
}


/*-----------------------------------------------
 * Get next move result
 *-----------------------------------------------
 * Returns FALSE if no more moves are possible.
 * psVal is expected to be the same in the whole
 * loop, so the counter is copied only the first
 * time.
 *-----------------------------------------------
 */

tBOOL bFieldSttLoopNext (tpsFIELDSTT psStt, tpsFIELDVAL psVal)
{
    tBOOL bMove = FALSE;

    if (psStt->cnNextLoop < psStt->cnNextNum)
      {
        if (psStt->cnNextLoop == 0)
          {
            vWordCntCopy(psVal->psCounter,cn_FieldVal_CntAll,psStt->psCounter,cn_FieldVal_CntAll);
            vWordCntCopy(psVal->psCounter,cn_FieldVal_CntNrm,psStt->psCounter,cn_FieldVal_CntNrm);
          }

        psVal->ulField = psStt->aulNextField[psStt->cnNextLoop];
        psVal->ulBit32 = psStt->aulNextBit32[psStt->cnNextLoop];
        psVal->cnMoves = psStt->acnNextMoves[psStt->cnNextLoop];

        psStt->cnNextLoop++;

        bMove = TRUE;
      }

    return (bMove);
}


/*-----------------------------------------------
 * Readout values for statistics
 *-----------------------------------------------
 */


tCOUNT cnFieldSttGetMovesAll (tpsFIELDSTT psStt)
{
    return (psStt->cnMovesAll);
}

/*---------------------------------------------*/

tCOUNT cnFieldSttGetMovesNrm (tpsFIELDSTT psStt)
{
    return (psStt->cnNextNum);
}

/*---------------------------------------------*/

tCOUNT cnFieldSttGetStones (tpsFIELDSTT psStt)
{
    return (psStt->cnStones);
}

/*---------------------------------------------*/

tCOUNT cnFieldSttGetSymType (tpsFIELDSTT psStt)
{
    return (psStt->cnSymType);
}

/*---------------------------------------------*/

tCOUNT cnFieldSttGetSymIndex (tpsFIELDSTT psStt)
{
    return (psStt->cnSymIndex);
}

/*---------------------------------------------*/

tBOOL bFieldSttGetIsWinning (tpsFIELDSTT psStt)
{
    return (psStt->bIsWinning);
}


/*-----------------------------------------------
 * Add Stt-counters to an external counter
 *-----------------------------------------------
 */


tVOID vFieldSttAddCountAll (tpsFIELDSTT psStt, tpsWORDCNT psCount, tCOUNT cnIndex)
{
    vWordCntAdd(psCount,cnIndex,psStt->psCounter,cn_FieldVal_CntAll);
}

/*---------------------------------------------*/

tVOID vFieldSttAddCountNrm (tpsFIELDSTT psStt, tpsWORDCNT psCount, tCOUNT cnIndex)
{
    vWordCntAdd(psCount,cnIndex,psStt->psCounter,cn_FieldVal_CntNrm);
}


/*----------------------------------------------------------------------*/
