

/* hashtabs.c */


/*------------------------------------------------------------------------
 * Module for management of the complete hash lookup tables
 *------------------------------------------------------------------------
 * This version of hashtabs is part of the Solitaire Brute Force tool.
 * Copyright (C) 2001 Holger Hasselbach <H.Hasselbach@galad.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *------------------------------------------------------------------------
 */


/*----------------------------------------------------------------------*/

#include "hashtabs.h"
#include "mymemory.h"
#include "hashblck.h"

#include <assert.h>

/*----------------------------------------------------------------------*/


struct s_HASHTABS
{
    tpsHASHBLCK apsLists[cnFieldHashRange];

    tpsHASHBLCK psLoopList;
};

typedef struct s_HASHTABS tsHASHTABS;


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Memory management open
 *-----------------------------------------------
 */

tVOID vHashTabsOpen (tVOID)
{
    vHashBlckOpen();
}


/*-----------------------------------------------
 * Memory management close
 *-----------------------------------------------
 */

tVOID vHashTabsClose (tVOID)
{
    vHashBlckClose();
}


/*-----------------------------------------------
 * Memory management statistics
 *-----------------------------------------------
 */

tVOID vHashTabsMemEmpty (tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    vHashBlckMemEmpty(pcnSizeof,pcnBlocks);
}


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Create a new HASHTABS
 *-----------------------------------------------
 */

tpsHASHTABS psHashTabsCreate (tVOID)
{
    tpsHASHTABS psTabs;
    tCOUNT      cnI;

    psTabs = pvMemoryAllocObject(sHASHTABS);

    for (cnI = 0; cnI < cnFieldHashRange; cnI++)
        psTabs->apsLists[cnI] = NULL;

    psTabs->psLoopList = NULL;

    return (psTabs);
}


/*-----------------------------------------------
 * Destroy a HASHTABS
 *-----------------------------------------------
 */

tpsHASHTABS psHashTabsDestroy (tpsHASHTABS psTabs)
{
    tCOUNT cnI;

    if (psTabs)
      {
        for (cnI = 0; cnI < cnFieldHashRange; cnI++)
            ppsHashBlckListDestroy(&(psTabs->apsLists[cnI]));

        ppsHashBlckListDestroy(&(psTabs->psLoopList));

        pvMemoryFree(psTabs);
      }

    return (NULL);
}


/*-----------------------------------------------
 * Memory statistics
 *-----------------------------------------------
 */

tpsHASHTABS psHashTabsMemUsage (tpsHASHTABS psTabs, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    tCOUNT cnI;

    if (psTabs)
      {
        for (cnI = 0; cnI < cnFieldHashRange; cnI++)
            ppsHashBlckListMemUsage(&(psTabs->apsLists[cnI]),pcnSizeof,pcnBlocks);

        ppsHashBlckListMemUsage(&(psTabs->psLoopList),pcnSizeof,pcnBlocks);

        *pcnSizeof += cnMemorySizeofObject(sHASHTABS);
        *pcnBlocks += 1;
      }

    return (psTabs);
}


/*-----------------------------------------------
 * Insert field
 *-----------------------------------------------
 */

tVOID vHashTabsInsert (tpsHASHTABS psTabs, tpsFIELDVAL psVal)
{
    tCOUNT cnHashKey;

    cnHashKey = cnFieldValHashKey(psVal);
    assert(cnHashKey < cnFieldHashRange);

    vHashBlckListInsert( &(psTabs->apsLists[cnHashKey]) ,psVal);
}


/*-----------------------------------------------
 * Initialize field loop
 *-----------------------------------------------
 */

tVOID vHashTabsLoopInit (tpsHASHTABS psTabs)
{
    tCOUNT       cnI;
    tppsHASHBLCK ppsBlck;

    ppsBlck = psTabs->apsLists;
    for (cnI = 0; cnI < cnFieldHashRange; cnI++, ppsBlck++)
        vHashBlckListConcat(&(psTabs->psLoopList),ppsBlck);
}


/*-----------------------------------------------
 * Get the next field
 *-----------------------------------------------
 * Returns FALSE if there are no more fields.
 *-----------------------------------------------
 */

tBOOL bHashTabsLoopNext (tpsHASHTABS psTabs, tpsFIELDVAL psVal)
{
    return ( bHashBlckListLoop(&(psTabs->psLoopList),psVal) );
}


/*-----------------------------------------------
 * Count the whole number of entries in the lists
 *-----------------------------------------------
 */

tCOUNT cnHashTabsEntriesList (tpsHASHTABS psTabs)
{
    tCOUNT       cnI,cnEntries = 0;
    tppsHASHBLCK ppsBlck;

    ppsBlck = psTabs->apsLists;
    for (cnI = 0; cnI < cnFieldHashRange; cnI++, ppsBlck++)
        cnEntries += cnHashBlckListEntries(ppsBlck);

    return (cnEntries);
}


/*-----------------------------------------------
 * Count the number of entries in a specific list
 *-----------------------------------------------
 */

tCOUNT cnHashTabsEntriesHash (tpsHASHTABS psTabs, tCOUNT cnHashKey)
{
    assert(cnHashKey < cnFieldHashRange);

    return ( cnHashBlckListEntries(&(psTabs->apsLists[cnHashKey])) );
}


/*-----------------------------------------------
 * Count the number of entries in the loop list
 *-----------------------------------------------
 */

tCOUNT cnHashTabsEntriesLoop (tpsHASHTABS psTabs)
{
    return ( cnHashBlckListEntries(&(psTabs->psLoopList)) );
}


/*----------------------------------------------------------------------*/
