

/* mystring.c */


/*------------------------------------------------------------------------
 * my-library with string functions for the tPSZ type
 *------------------------------------------------------------------------
 * This version of mystring is part of the Solitaire Brute Force tool.
 * Copyright (C) 2001 Holger Hasselbach <H.Hasselbach@galad.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *------------------------------------------------------------------------
 */


/*----------------------------------------------------------------------*/

#include "mystring.h"
#include "mymemory.h"

#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <assert.h>

/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Get string length
 *-----------------------------------------------
 * The same as strlen, but NULL is valid and returns 0.
 *-----------------------------------------------
 */

tCOUNT cnStringLength (tPSZc pzStr)
{
    return ( pzStr ? strlen(pzStr) : 0 );
}


/*-----------------------------------------------
 * Substitute characters in string
 *-----------------------------------------------
 * It is possible (but not recommended) to substitute
 * a character to '\0'.
 * NULL is valid for pzStr.
 *-----------------------------------------------
 */

tPSZ pzStringSubstChar (tPSZ pzStr, tCHAR cOld, tCHAR cNew)
{
    tCOUNT cnLen,cnI;

    cnLen = cnStringLength(pzStr);
    for (cnI = 0; cnI < cnLen; cnI++)
        if (pzStr[cnI] == cOld)
            pzStr[cnI] = cNew;

    return (pzStr);
}


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Allocate memory for string
 *-----------------------------------------------
 * cnLength is exclusive the closing '\0', meaning
 * the same as cnStringLength returns.
 *-----------------------------------------------
 */

tpPSZ ppzStringReserve (tpPSZ ppzStr, tCOUNT cnLength)
{
    ppzStringClear(ppzStr);

    if (cnLength)
      {
        *ppzStr = pvMemoryAlloc(sizeof(**ppzStr) * (cnLength + 1));
        (*ppzStr)[cnLength] = '\0';
      }

    return (ppzStr);
}


/*-----------------------------------------------
 * Free memory of string
 *-----------------------------------------------
 */

tpPSZ ppzStringClear (tpPSZ ppzStr)
{
    if (*ppzStr)
        *ppzStr = pvMemoryFree(*ppzStr);

    return (ppzStr);
}


/*-----------------------------------------------
 * Memory statistics
 *-----------------------------------------------
 */

tpPSZ ppzStringMemUsage (tpPSZ ppzStr, tpCOUNT pcnSizeof, tpCOUNT pcnBlocks)
{
    if (*ppzStr)
      {
        *pcnSizeof += sizeof(**ppzStr) * (cnStringLength(*ppzStr) + 1);
        *pcnBlocks += 1;
      }

    return (ppzStr);
}


/*-----------------------------------------------
 * Copy a string
 *-----------------------------------------------
 */

tpPSZ ppzStringCopy (tpPSZ ppzStr, tPSZc pzSrc)
{
    tPSZ   pzNew = NULL;
    tCOUNT cnLength;

    cnLength = cnStringLength(pzSrc);
    if (cnLength)
      {
        ppzStringReserve(&pzNew,cnLength);
        memcpy(pzNew,pzSrc,sizeof(*pzNew) * cnLength);
      }

    ppzStringClear(ppzStr);
    *ppzStr = pzNew;

    return (ppzStr);
}


/*-----------------------------------------------
 * Concatenate any number of strings
 *-----------------------------------------------
 * NULL is always valid for the source strings, so
 * it can't be used as an end flag.
 * The correctness of cnNumber is at the mercy of
 * the programmer, thus can't be checked anyway.
 * cnNumber==0 is valid.
 *-----------------------------------------------
 */

tpPSZ ppzStringConcatN (tpPSZ ppzStr, tCOUNT cnNumber, ...)
{
    tCOUNT  cnLength,cnIndex,cnI,cnSrcLen;
    tPSZ    pzSrc,pzNew = NULL;
    va_list vaArgs;

    va_start(vaArgs,cnNumber);
    cnLength = 0;
    for (cnI = 0; cnI < cnNumber; cnI++)
      {
        pzSrc = va_arg(vaArgs,tPSZ);
        cnLength += cnStringLength(pzSrc);
      }
    va_end(vaArgs);

    if (cnLength)
      {
        ppzStringReserve(&pzNew,cnLength);

        va_start(vaArgs,cnNumber);
        cnIndex = 0;
        for (cnI = 0; cnI < cnNumber; cnI++)
          {
            pzSrc = va_arg(vaArgs,tPSZ);
            cnSrcLen = cnStringLength(pzSrc);
            if (cnSrcLen)
              {
                memcpy(&(pzNew[cnIndex]),pzSrc,sizeof(*pzNew) * cnSrcLen);
                cnIndex += cnSrcLen;
              }
          }
        va_end(vaArgs);
      }

    ppzStringClear(ppzStr);
    *ppzStr = pzNew;

    return (ppzStr);
}


/*-----------------------------------------------
 * Copy string parts
 *-----------------------------------------------
 * cnIndex : Index in pzSrc (0..cnStringLength(pzSrc)-1)
 * cnNumber: Number of characters to copy
 *-----------------------------------------------
 */

tpPSZ ppzStringMiddle (tpPSZ ppzStr, tPSZc pzSrc, tCOUNT cnIndex, tCOUNT cnNumber)
{
    tPSZ   pzNew = NULL;
    tCOUNT cnLength;

    cnLength = cnStringLength(pzSrc);
    if ( cnNumber && (cnIndex < cnLength) )
      {
        if (cnNumber > (cnLength - cnIndex))
            cnNumber = cnLength - cnIndex;

        ppzStringReserve(&pzNew,cnNumber);
        memcpy(pzNew,&(pzSrc[cnIndex]),sizeof(*pzNew) * cnNumber);
      }

    ppzStringClear(ppzStr);
    *ppzStr = pzNew;

    return (ppzStr);
}


/*-----------------------------------------------
 * Build a string with a multiple character
 *-----------------------------------------------
 */

tpPSZ ppzStringMultiChar (tpPSZ ppzStr, tCHAR cChar, tCOUNT cnNumber)
{
    ppzStringReserve(ppzStr,cnNumber);

    while (cnNumber--)
        (*ppzStr)[cnNumber] = cChar;

    return (ppzStr);
}


/*----------------------------------------------------------------------*/


/*
 * How to convert a big integer of arbitrary length (means: arbitrary number
 * of bits) into an ASCII string to an arbitrary number system base without
 * using big integer division and modulus:
 *
 * Basically, there are two ways to do the conversion.
 *
 * a) For each digit of the target string, beginning with least significant
 *        digit = integer % base
 *        integer = integer / base
 *    Until integer == 0
 *
 * b) For each 1-bit of the integer in any order
 *        Add value of bit in wanted number system to the target string
 *
 * Obviously, we want b). The value of bit bi in an integer with Nb bits and
 * i from 0 to Nb-1 is: bi * 2^i. The value of the whole integer is:
 *   bNb-1 * 2^Nb-1 + bNb-2 * 2^Nb-2 + ... + b1 * 2^1 + b0 * 2^0
 * = (...((bNb-1)*2 + bNb-2)*2 + ... + b1)*2 + b0
 *
 * b) Initialize target string with 0
 *    For each bit of the integer, beginning with most significant
 *        Initialize carry with bit
 *        digits = digits * 2, calculate carry for wanted number system
 *
 * The string could be initialized with bNb-1 and the loop could begin with
 * bNb-2, but in the nested code loops this would cause a special case.
 * The first used version:
 *        digits = digits * 2
 *        digits[0] += bit
 * didn't work for odd bases. Ooops.
 *
 *
 * What amount of characters is necessary for the string?
 *
 * An integer with Nb bits has a maximum value of 2^Nb - 1. A number in a
 * number system to the base d with Nd digits has a maximum value of
 * d^Nd - 1. For the d-number to be able to express the full integer range,
 * it is necessary that
 *     d^Nd - 1 >= 2^Nb - 1
 * ==>     d^Nd >= 2^Nb
 * ==>       Nd >= logd(2^Nb)    ; logd = logarithm to base d
 * ==>       Nd >= Nb * logd(2)
 *
 * This works for any conversion between number systems. Let d be the target
 * base and Ld(x) the logarithm of x to the base d. To convert a number given
 * in the base b with Nb digits into a system to base d, Nd digits are
 * necessary:
 *
 * Nd = Nb * Ld(b)
 *
 * For a 16 bit integer to be expressed in decimal, Nd needs to be at least
 * 16 * log10(2) = 16 * 0.30103 = 4.82. Since Nd itself is an integer, the
 * next full value 5 is the size of choice for 65535. Don't forget one
 * additional character for the string closing \0.
 * With this, the full formular is:
 *
 * Nd = ceil( Nb * Ld(b) ) + 1
 *
 * Can you calculate the logarithm of 2 to the base 7 on your pocket
 * calculator? I can, because
 *
 * Ld(b) = La(b) / La(d) with a = any base, for the calculator and for C
 *                       functions 10 or e.
 *
 * So, L7(2) = L10(2) / L10(7) = 0.35621
 *
 * Here are the factors Ld(2) to convert a binary integer into any base d
 * from 2 to 16. I didn't find any practical use for systems above 16. The
 * next useful as far as I see would be 256... For the sake of the correct
 * string size, the factors are always rounded upward.
 */


static tDOUBLE afd_String_NumFactorsLd[uwStringNumMaxBase] =
                 {
                   0.00000, /*  1 (index 0) */
                   1.00000, /*  2           */
                   0.63093, /*  3           */
                   0.50000, /*  4           */
                   0.43068, /*  5           */
                   0.38686, /*  6           */
                   0.35621, /*  7           */
                   0.33334, /*  8           */
                   0.31547, /*  9           */
                   0.30103, /* 10           */
                   0.28907, /* 11           */
                   0.27895, /* 12           */
                   0.27024, /* 13           */
                   0.26265, /* 14           */
                   0.25596, /* 15           */
                   0.25000  /* 16           */
                 };


static tCHAR ac_String_NumChars[uwStringNumMaxBase] =
               {
                 '0','1','2','3','4','5','6','7',
                 '8','9','a','b','c','d','e','f'
               };


/*-----------------------------------------------
 * Get the string size for conversion to base
 *-----------------------------------------------
 * Nd = ceil( Nb * Ld(b) ) + 1
 * +1 for the string closing \0 (thus Size, not Length)
 * If Words is 0, return is 2 for "0".
 *-----------------------------------------------
 */

tCOUNT cnStringNumStrSize (tCOUNT cnWords, tUWORD uwBase)
{
    assert( (uwBase >= 2) && (uwBase <= uwStringNumMaxBase) );

    return ( cnWords
             ? (tCOUNT)ceil( (tDOUBLE)(cnWords * 16) * afd_String_NumFactorsLd[uwBase - 1] ) + 1
             : 2
           );
}


/*-----------------------------------------------
 * Convert WORD-array to a number string
 *-----------------------------------------------
 * pzStr is expected to have the right size
 * including the closing \0.
 * WORD-array is expected to be unsigned.
 * If Words is 0, return is "0".
 *-----------------------------------------------
 */

tPSZ pzStringNumIntToStr (tPSZ pzStr, tpcUWORD puwInt, tCOUNT cnWords, tUWORD uwBase)
{
    tCOUNT  cnBit,cnDigit;
    tCHAR   cAdd;
    tCOUNT  cnLength;

    assert( (uwBase >= 2) && (uwBase <= uwStringNumMaxBase) );

    pzStr[0] = 0;
    cnLength = 1;

    puwInt += cnWords;
    while (cnWords--)
      {
        puwInt--;
        cnBit = 16;
        while (cnBit--)
          {
            cAdd = (tCHAR)((*puwInt >> cnBit) & 1);
            for (cnDigit = 0; cnDigit < cnLength; cnDigit++)
              {
                cAdd += pzStr[cnDigit] * 2;
                if (cAdd >= (tCHAR)uwBase)
                  {
                    pzStr[cnDigit] = cAdd - (tCHAR)uwBase;
                    cAdd = 1;
                  }
                else
                  {
                    pzStr[cnDigit] = cAdd;
                    cAdd = 0;
                  }
              }

            if (cAdd)
                pzStr[cnLength++] = cAdd;
          }
      }

    for (cnDigit = 0; cnDigit < cnLength; cnDigit++)
        pzStr[cnDigit] = ac_String_NumChars[ (tCOUNT)pzStr[cnDigit] ];

    for (cnDigit = 0; cnDigit < (cnLength >> 1); cnDigit++)
      {
        cAdd = pzStr[cnDigit];
        pzStr[cnDigit] = pzStr[cnLength - 1 - cnDigit];
        pzStr[cnLength - 1 - cnDigit] = cAdd;
      }

    pzStr[cnLength] = '\0';

    return (pzStr);
}


/*-----------------------------------------------
 * Convert UInt WORD-array to string
 *-----------------------------------------------
 * ppzStr is dynamically allocated to the right size.
 *-----------------------------------------------
 */

tpPSZ ppzStringUIntToStr (tpPSZ ppzStr, tpcUWORD puwInt, tCOUNT cnWords, tUWORD uwBase)
{
    ppzStringReserve(ppzStr, cnStringNumStrSize(cnWords,uwBase) - 1);
    pzStringNumIntToStr(*ppzStr,puwInt,cnWords,uwBase);

    return (ppzStr);
}


/*-----------------------------------------------
 * Convert ULONG to string
 *-----------------------------------------------
 * ppzStr is dynamically allocated to the right size.
 *-----------------------------------------------
 */

tpPSZ ppzStringULongToStr (tpPSZ ppzStr, tULONG ulInt, tUWORD uwBase)
{
    tUWORD auwInt[2];

    auwInt[0] = (tUWORD)(ulInt & 0xffff);
    auwInt[1] = (tUWORD)( (ulInt >> 16) & 0xffff );

    return ( ppzStringUIntToStr(ppzStr,auwInt,2,uwBase) );
}


/*-----------------------------------------------
 * Convert SLONG to string
 *-----------------------------------------------
 * ppzStr is dynamically allocated to the right size
 * including the '-'.
 *-----------------------------------------------
 */

tpPSZ ppzStringSLongToStr (tpPSZ ppzStr, tSLONG slInt, tUWORD uwBase)
{
    if (slInt < 0)
      {
        ppzStringULongToStr(ppzStr,-slInt,uwBase);
        ppzStringConcatN(ppzStr,2,"-",*ppzStr);
      }
    else
        ppzStringULongToStr(ppzStr,slInt,uwBase);

    return (ppzStr);
}


/*----------------------------------------------------------------------*/


/*-----------------------------------------------
 * Format a number string
 *-----------------------------------------------
 * The formating is done in the order of the parameters.
 * cnZeroFill: If the length of the string is less than
 *             this, the string is filled up with 0 on
 *             the left side.
 *             "a14bcde" --> 11 --> "0000a14bcde"
 * cnBlkSize : If this is != 0, the string is separated
 *             into blocks beginning from the right.
 * pzBlkSep  : This defines the block separation.
 *             "0000a14bcde" --> 4,":" --> "000:0a14:bcde"
 * pzPrefix  : At last, Prefix and
 * pzSuffix  : Suffix are concatenated to the string.
 *             "000:0a14:bcde" --> "0x","" --> "0x000:0a14:bcde"
 * The string is formatted 'as is', it is not changed before,
 * and there is no check for a sign '-' or '+'.
 *-----------------------------------------------
 */

tpPSZ ppzStringFormatNumber (tpPSZ ppzStr, tPSZc pzNumber, tCOUNT cnZeroFill, tCOUNT cnBlkSize,
                             tPSZc pzBlkSep, tPSZc pzPrefix, tPSZc pzSuffix)
{
    tCOUNT cnLenNumber,cnLenBlkSep,cnLenPrefix,cnLenSuffix,
           cnLenZero,cnNumSep,cnLenNew,
           cnIndNew,cnI;
    tPSZ   pzNew = NULL;

    cnLenNumber = cnStringLength(pzNumber);
    cnLenBlkSep = cnStringLength(pzBlkSep);
    cnLenPrefix = cnStringLength(pzPrefix);
    cnLenSuffix = cnStringLength(pzSuffix);

    cnLenZero = ( cnZeroFill > cnLenNumber ? cnZeroFill : cnLenNumber );
    cnNumSep  = ( cnBlkSize && cnLenZero ? (cnLenZero - 1) / cnBlkSize : 0 );

    cnLenNew = cnLenPrefix + cnLenZero + (cnNumSep * cnLenBlkSep) + cnLenSuffix;
    if (cnLenNew)
      {
        ppzStringReserve(&pzNew,cnLenNew);
        cnIndNew = 0;

        while (cnLenPrefix--)
            pzNew[cnIndNew++] = *(pzPrefix++);

        while (cnLenZero)
          {
            pzNew[cnIndNew++] = ( cnLenZero > cnLenNumber ? '0' : *(pzNumber++) );
            cnLenZero--;

            if ( cnNumSep && cnLenZero && ((cnLenZero % cnBlkSize) == 0) )
              {
                for (cnI = 0; cnI < cnLenBlkSep; cnI++)
                    pzNew[cnIndNew++] = pzBlkSep[cnI];
                cnNumSep--;
              }
          }

        while (cnLenSuffix--)
            pzNew[cnIndNew++] = *(pzSuffix++);
      }

    ppzStringClear(ppzStr);
    *ppzStr = pzNew;

    return (ppzStr);
}


/*----------------------------------------------------------------------*/
